"""
Symbolic Regression Toolkit

This package provides a Python-based toolkit for equation discovery/symbolic regression.

Modules:
    utils: The module containing utility classes and functions.
    evaluation: The module containing classes and functions for estimating parameters and evaluating Symbolic Regression models.
"""

__version__ = "1.2.2"
__author__ = "Sebastian Mežnar"
__credits__ = "Jožef Stefan Institute"

from . import utils
from . import evaluation
from . import dataset